@@ -1,37 +1,28 @@
 /*
  * BluezQt - Asynchronous BlueZ wrapper library
  *
- * Copyright (C) 2014-2015 David Rosca <nowrep@gmail.com>
+ * SPDX-FileCopyrightText: 2014-2015 David Rosca <nowrep@gmail.com>
  *
- * This library is free software; you can redistribute it and/or
- * modify it under the terms of the GNU Lesser General Public
- * License as published by the Free Software Foundation; either
- * version 2.1 of the License, or (at your option) version 3, or any
- * later version accepted by the membership of KDE e.V. (or its
- * successor approved by the membership of KDE e.V.), which shall
- * act as a proxy defined in Section 6 of version 3 of the license.
- *
- * This library is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
- * Lesser General Public License for more details.
- *
- * You should have received a copy of the GNU Lesser General Public
- * License along with this library. If not, see <http://www.gnu.org/licenses/>.
+ * SPDX-License-Identifier: LGPL-2.1-only OR LGPL-3.0-only OR LicenseRef-KDE-Accepted-LGPL
  */
 
 #ifndef BLUEZQT_PENDINGCALL_H
 #define BLUEZQT_PENDINGCALL_H
 
+#include <functional>
+
 #include <QObject>
 
 #include "bluezqt_export.h"
 
+#include <memory>
+
+class QDBusError;
 class QDBusPendingCall;
+class QDBusPendingCallWatcher;
 
 namespace BluezQt
 {
-
 /**
  * @class BluezQt::PendingCall pendingcall.h <BluezQt/PendingCall>
  *
@@ -96,19 +87,23 @@ public:
         AuthenticationTimeout = 18,
         /** Indicates that the connection attempt have failed. */
         ConnectionAttemptFailed = 19,
+        /** Indicates that the data provided generates a data packet which is too long. */
+        InvalidLength = 20,
+        /** Indicates that the action is not permitted (e.g. maximum reached or socket locked). */
+        NotPermitted = 21,
         /** Indicates an error with D-Bus. */
         DBusError = 98,
         /** Indicates an internal error. */
         InternalError = 99,
         /** Indicates an unknown error. */
-        UnknownError = 100
+        UnknownError = 100,
     };
     Q_ENUM(Error)
 
     /**
      * Destroys a PendingCall object.
      */
-    ~PendingCall();
+    ~PendingCall() override;
 
     /**
      * Returns a first return value of the call.
@@ -178,26 +173,41 @@ private:
         ReturnVoid,
         ReturnUint32,
         ReturnString,
+        ReturnStringList,
         ReturnObjectPath,
         ReturnFileTransferList,
-        ReturnTransferWithProperties
+        ReturnTransferWithProperties,
+        ReturnByteArray
     };
 
-    explicit PendingCall(const QDBusPendingCall &call, ReturnType type, QObject *parent = nullptr);
-    explicit PendingCall(Error error, const QString &errorText, QObject *parent = nullptr);
+    BLUEZQT_NO_EXPORT explicit PendingCall(Error error, const QString &errorText, QObject *parent = nullptr);
+    BLUEZQT_NO_EXPORT explicit PendingCall(const QDBusPendingCall &call, ReturnType type, QObject *parent = nullptr);
 
-    class PendingCallPrivate *const d;
+    // exported because called from template BluezQt::TPendingCall constructor
+    using ErrorProcessor = std::function<void(const QDBusError &error)>;
+    using ExternalProcessor = std::function<void(QDBusPendingCallWatcher *watcher, ErrorProcessor errorProcessor, QVariantList *values)>;
+    explicit PendingCall(const QDBusPendingCall &call, ExternalProcessor externalProcessor, QObject *parent = nullptr);
+
+    std::unique_ptr<class PendingCallPrivate> const d;
 
     friend class PendingCallPrivate;
     friend class Manager;
     friend class Adapter;
+    friend class GattServiceRemote;
+    friend class GattCharacteristicRemote;
+    friend class GattDescriptorRemote;
     friend class Device;
+    friend class GattManager;
+    friend class LEAdvertisingManager;
+    friend class Media;
     friend class MediaPlayer;
     friend class ObexManager;
     friend class ObexTransfer;
     friend class ObexSession;
     friend class ObexObjectPush;
     friend class ObexFileTransfer;
+    template<class... T>
+    friend class TPendingCall;
 };
 
 } // namespace BluezQt
